//
// J1850 protocol converter
// Controller Freescale MC9S12DJ64 and interface chip MC33390
// Compiler CodeWarrior IDE version 5.7.0
//
// Juha Niinikoski 22.05.2007
//
// SCI0 IRQ driver 24.05.2007
// first read of meaningfull data 25.05.2007
// VIN number request & raw display 27.05.2007
// BDLC interrupt promoted with HPRIO 30.05.2007
// more test code, display data filter 30.05.2007
// LCD display speed optimization & Jeep protocol displays 31.05.2007
// more Jeep protocols discovered 01.06.2007
// more Jeep protocols discovered 21.06.2007
// LCD back light intensity with pwm output 26.06.2007
// LCD back light control from bus message 07.07.2007

#include <hidef.h>          /* common defines and macros */
#include <mc9s12dj64.h>     /* derivative information */
#pragma LINK_INFO DERIVATIVE "mc9s12dj64"

#include <stdio.h>
#include "J1850_conv.h"                     // board specific definitions
#include "lcd.h"                            // LCD module definitions
#include "uart0_irq.h"                      // SCI0 module definitions



// not in  mc9s12dj64.h file
/*** DLCBSTAT - BDLC Status Register; 0x000000EF ***/

volatile unsigned char DLCBSTAT @(REG_BASE + 0x000000EF);


// Global data

const char ver[] = {"07.07.07"};            // test software version
const char msg1[] = {"J1850   "};
const char msg2[] = {"Juha Niinikoski "};

// Timing, global variables

static unsigned long absoluteTime = 0;
static unsigned int ms_timer;                        // ms delay timer, count from set value to zero


// Diag message constants / data

// J1979 constants
#define MODE1 0x01
#define MODE2 0x02
#define MODE9 0x09

// some test data requests
// request Temp, IAT, Fuel syst stat,TPS, RPM
#define REQUEST_IDX 5       // number of messages
#define REQUEST_TIMER 200   // try 200 ms poll intervall


char request_idx = 0;
//const unsigned char requests[REQUEST_IDX] = {0x05, 0x0F, 0x03, 0x11, 0x0C}; // Temp, IAT, Stat, TPS, RPM
//const unsigned char requests[REQUEST_IDX] = {0x05, 0x0F, 0x03, 0x0B, 0x0C};  // Temp, IAT, Stat, MAP, RPM
const unsigned char requests[REQUEST_IDX] = {0x05, 0x0F, 0x03, 0x0B, 0x0E};  // Temp, IAT, Stat, MAP, ADV

#define DISP_PAGES 9                        // number of display pages
char disp_page = 0;                         // selected LCD display page
unsigned char lcd_pwm = 128;                // lcd brightness

// VIN request data
char vin_packet;                            // number of received VIN responce packets
char vin_string[18];                        // collected VIN
unsigned char vin_ecu;                      // answer from module #?


// J1850 controller data

#define MY_ADDRESS 0xF0                     // my J1850 address

// transmit, output buffer
#define J1850_TX_BUF_LEN 16
volatile unsigned char j1850_tx_buf[J1850_TX_BUF_LEN];
char j1850_tx_bytes = 0;
char tdre_cnt = 0;                           // bytes transmitted

// BDLC flags
char loa_flag= 0;                           // los of arbitration flag
char rc4x_flag = 0;                         // 4x mode flag
char retx_flag = 0;                         // retransmit flag
char eof = 0;                               // 1 = eof received, >1 = buffer overrun      

// receive, input buffer
#define J1850_RX_BUF_LEN 16
volatile unsigned char j1850_rx_buf[J1850_RX_BUF_LEN];
char j1850_rx_bytes = 0;

// receive display/output buffer
#define J1850_DISP_BUF_LEN 16                 
volatile unsigned char j1850_disp_buf[J1850_DISP_BUF_LEN];
char j1850_disp_bytes = 0;

// display filter
#define DISP_FILTER_LEN 4
volatile unsigned char disp_filter[DISP_FILTER_LEN] = {0x48,0,0,0}; // display data filter, init for diag messages
char disp_filt_digit = 0;

// BDLC staristics
unsigned long rx_counter = 0;               // RX messages
unsigned int crc_errors = 0;                // CRC errors, counter
unsigned int ifr_messages = 0;              // IFR flags
unsigned int ivalid_symbols = 0;            // INVALID symbols
unsigned int rx_overflow = 0;               // RX buffer overflow bytes

unsigned int tx_counter = 0;                // TX messages
unsigned int loa_counter = 0;               // LOA situations

unsigned int drop_frames = 0;               // drop because previous frame still being processed



// BDLC interrupt code

void bdlc_handler(void)                    // BDLC operation selector
  {
  unsigned char dummy;
  char x;
  
   switch (DLCBSVR)                        // select operation with BSVR vector 
			{
			case 0x00: default:                  // BDLC clear, no interrupts
			asm ("nop");                         // dummy
      break;
      
// EOF flag      
 			case 0x04:                          // Received EOF, end of frame
			if(loa_flag != 0)
			  loa_flag = 0;                     // clear loa flag
			else
			  if(tdre_cnt != 0)                 // EOF after TX
			    {
			    tdre_cnt = 0;                   // clear TX queue, all done
			    j1850_tx_bytes = 0; 
			    }
			    
			rx_counter++;                       // do statistics 

      if(eof == 0)                        // buffer used / free for new data   			
          {  
			    j1850_disp_bytes = j1850_rx_bytes;  // copy message to display buffer
			    j1850_rx_bytes = 0;                 // reset input fuffer pointer  
			    
			    for(x=0; x < j1850_disp_bytes; x++) // copy buffer
			      {
			      j1850_disp_buf[x] = j1850_rx_buf[x];  
			      }

			    eof = 1;                        // eof occurred, set flag 
          }
           
      else
        j1850_rx_bytes = 0;                 // reset input fuffer pointer     

      break;
      
// IFR flag      
 			case 0x08:                          // Received IFR byte
			ifr_messages++;                     // do statistics 
      break;
      
// RX data
 			case 0x0C:                          // Rx data register full
 			  if(j1850_rx_bytes < 15)           // test, buffer limit watch
 			      {
			      j1850_rx_buf[j1850_rx_bytes] = DLCBDR; // save data
			      j1850_rx_bytes++;            // prepere for next byte
 			      }
 			   else
 			      { 
 			      dummy = DLCBDR;               // carbage data
 			      rx_overflow++;                // do statistics
 			      }
      break;
      
// TX data
      case 0x10:                           // Tx data register empty
			if( tdre_cnt == (j1850_tx_bytes) )
			   {
			   DLCBCR2_TEOD = 1;                 // set TEOD bit, last byte going out
			   tx_counter++;                     // do statistics
			   }
			else
			  {   
			  DLCBDR = j1850_tx_buf[tdre_cnt];   // load tx register
			  tdre_cnt++;   
			  }
      break;
      
// LOA, Loss of arbitration
      case 0x14:                          // Loss of arbitratio
			loa_flag++;                         // increment loss of arbitration
			tdre_cnt = 0;                       // clear transmitted count
			loa_counter++;                      // keep statistics 
      break;
      
// CRC error      
      case 0x18:                          // CRC error
			if(tdre_cnt != 0)                   // check if occurred during transmission
			  {
			  tdre_cnt = 0;                     // clear transmitted count, retransmit
			  retx_flag++;                      // increment retransmit flag
			  }
			else
			  j1850_rx_bytes = 0;               // clear RX byte count
			  
			 crc_errors++;                      // increment error counter  
      break;
      
// General BUS error      
      case 0x1C:                          // Symbol Invalid or out of range
			if(tdre_cnt != 0)                   // check if occurred during transmission
			  {
			  tdre_cnt = 0;                     // clear transmitted count
			  retx_flag++;                      // increment retransmit flag
			  }
			  
			 else
			  {
			  j1850_rx_bytes = 0;               // clear RX byte count    
			  }
			  
			 ivalid_symbols++;                  // do statistics 
      break;

// Wake up      
      case 0x20:                          //  Wakeup
			asm ("nop");                         // dummy, we do not sleep 
      break;
                
		}   // end switch (DLCBSVR)
  }     // end void bdlc_handler(void)




// Interrupt code, keep it in right segment

// !!!!!!!! Do not forget to install IRQ vectors in xxxlinker.prm file !!!!!!!!

//VECTOR 0 _Startup /* reset vector: this is the default entry point for a C/C++ application. */

// J1850 converter application vectors. Juha Niinikoski 23.05.2007
//VECTOR 7  RTI_ISR
//VECTOR 20 SCI0_ISR
//VECTOR 30 BDLC_ISR


#pragma CODE_SEG __NEAR_SEG NON_BANKED
interrupt void RTI_ISR(void)
  {
  absoluteTime++;                         // tick counter
  
  if(ms_timer != 0)                       // do ms timer
    ms_timer--;
  
   CRGFLG = 0x80;                          // clear RTIF bit
  
//   LED_PORT = LED_PORT ^ LD2;              // test 
  }
  
interrupt void BDLC_ISR(void)
  {
  LED_PORT = LED_PORT | LD1;              // BDLC int indicator
  bdlc_handler();                         // run BDLC code
  LED_PORT = LED_PORT & ~LD1;
  }

interrupt void SCI0_ISR(void)
  {
  LED_PORT = LED_PORT | LD2;              // SCI0 int indicator
  sci0_int();                             // run SCI0 int code
  LED_PORT = LED_PORT & ~LD2;
  }
  
#pragma CODE_SEG DEFAULT




// start timer tick
static void RTIInit(void)
  {
  /* setup of the RTI interrupt frequency */
  /* adjusted to get 1 millisecond (1.024 ms) with 16 MHz oscillator */
  RTICTL = 0x1F;  /* set RTI prescaler */ 
  CRGINT = 0x80;  /* enable RTI interrupts */ 
  }



// BDLC controler init

void init_J1850(void)                     // MC9S12DJ64 BDLC module init
{
char temp;
char i=0xFF;

MC33390_CONTROL_PORT = MC33390_CONTROL_PORT |  MC33390_SLEEP; // wake up transceiver, keep loop back

DLCBARD = 0x47;                            // off chip transceiver,16us round trip

DLCBCR2 = 0xC0;                            //  set loop back
DLCBCR1 = 0x00;                            // Integer Freq. 8MHz,disable interrupts
  

DLCBRSR = 0x07;                            // divide BDLC clock from 8MHz E-cloc
DLCSCR = 0x10;                             // BDLC enable

DLCBCR2 = 0x40;                            // take bdlc out of loop back

while((i>0)^(temp!=0x00))                  // check BDLC to make sure no interrupts are waiing
  {
  temp=DLCBSVR;
  i--;
  }
  

MC33390_CONTROL_PORT = MC33390_CONTROL_PORT &~ MC33390_LOOP_BACK; // normal waveshaping, 4X/LOOP = 0
MC33390_LOOP_PORT = MC33390_LOOP_PORT | MC33390_LOOP_BACK;       // take transceiver out of loop back

DLCBCR2 = 0x00;                            //  remove loop back
DLCBCR1 = 0x02;                            // Integer Freq. 8MHz,enable BDLC interrupts 
}


// Print J1850 RX buffer to terminal

void print_j1850_rx(unsigned char amount)
  {
   char x;
   for(x=0; x<amount; x++)
      {
      putchhex(j1850_disp_buf[x]);
      putch(' '); 
      }
   putstr("\n\r");
  }

// Print BDLC registers, debug printout

void print_reg(void)
  {   
   putstr("\n\rBCR1 =");  
   putchhex(DLCBCR1); 
 
   putstr("\n\rBCR2 =");  
   putchhex(DLCBCR2);
   
   putstr("\n\rBSVR =");
   putchhex(DLCBSVR);
   
   putstr("\n\rBDR  =");
   putchhex(DLCBDR);
   
   putstr("\n\rBSTAT=");
   putchhex(DLCBSTAT);    

   putstr("\n\r_____________");   
  }




// Print collected J1850 bus statistics

void print_statistic(void)
  {
  putstr("\n\rJ1850 bus statistics");
  
  putstr("\n\r  RX counter = ");  
  putulong(rx_counter);
 
  putstr("\n\r  CRC errors = ");
  putuint(crc_errors);
  
  putstr("\n\rIFR messages = ");
  putuint(ifr_messages);
  
  putstr("\n\rInval symbol = ");
  putuint(ivalid_symbols);
  
  putstr("\n\rLong RX msgs = ");
  putuint(rx_overflow);

  putstr("\n\r  TX counter = ");
  putuint(tx_counter);

  putstr("\n\r LOA counter = ");
  putuint(loa_counter);
  
  putstr("\n\r Drop frames = ");
  putuint(drop_frames);
  
  putstr("\n\r  Time stamp = ");
//  putlonghex(absoluteTime);
  putulong(absoluteTime);
  
  putstr("\n\r********************");
  }



// diag routines

// J1850 send, buffer must be filled elsewhere
void J1850_send(unsigned char txb)
  {
  while(tdre_cnt != 0);                   // wait for free transmitter
        
        j1850_tx_bytes = txb;               // set message length 
//        tdre_cnt = 0;                     // prepare transmitter
        
        DLCBDR = j1850_tx_buf[tdre_cnt];  // load tx register, tx first byte
        tdre_cnt++;
  }

// J1979 send
void J1979_send(unsigned char mode, unsigned char pid)
  {
  while(tdre_cnt != 0);                   // wait for free transmitter
  
   j1850_tx_buf[0] = 0x68;                // set J1979 Priority/type
        j1850_tx_buf[1] = 0x6A;           // set logical target, Diag Request
        j1850_tx_buf[2] = MY_ADDRESS;     // use scan tool address 0xF1             
        j1850_tx_buf[3] = mode;           // set J1979 mode
        j1850_tx_buf[4] = pid;            // set J1979 PID
        
        j1850_tx_bytes = 5;               // set message length 
//        tdre_cnt = 0;                     // prepare transmitter
        
        DLCBDR = j1850_tx_buf[tdre_cnt];  // load tx register, tx first byte
        tdre_cnt++;
  }


// J1979 diagnostic command dispatcher
void send_diag_commands(void)
  {
  if(ms_timer == 0)
    {
    ms_timer = REQUEST_TIMER;                  // poll interval
      
     J1979_send(MODE1, requests[request_idx]); // send command
     
     request_idx++;                       // addjust index for next round
     if(request_idx == REQUEST_IDX)
        request_idx = 0;      
    }
  }


// Read VIN responce packets
char read_vin(void)
  {
   if(eof)
        {
        eof = 0;                                  // buffer used
        if((j1850_disp_buf[0] == 0x48) && (j1850_disp_buf[3] == 0x49)) // display filter, only diag MODE 09 responce messages 
            {
            vin_ecu = j1850_disp_buf[2];          // save responding ecu #
            switch(vin_packet)
              {
		    	    case 0:
	            vin_string[0] =  j1850_disp_buf[9]; // 1st digit
		    	    break;
		    	    
		    	    case 1:
	            vin_string[1] =  j1850_disp_buf[6]; // digits 2-5
	            vin_string[2] =  j1850_disp_buf[7];
	            vin_string[3] =  j1850_disp_buf[8];
	            vin_string[4] =  j1850_disp_buf[9];
		    	    break;

		    	    case 2:
	            vin_string[5] =  j1850_disp_buf[6]; // digits 6-9
	            vin_string[6] =  j1850_disp_buf[7];
	            vin_string[7] =  j1850_disp_buf[8];
	            vin_string[8] =  j1850_disp_buf[9];
		    	    break;
		    	    
		    	    case 3:
	            vin_string[9] =  j1850_disp_buf[6]; // digits 10-13
	            vin_string[10] =  j1850_disp_buf[7];
	            vin_string[11] =  j1850_disp_buf[8];
	            vin_string[12] =  j1850_disp_buf[9];
		    	    break;
		    	    
		    	    case 4:
	            vin_string[13] =  j1850_disp_buf[6]; // digits 14-17
	            vin_string[14] =  j1850_disp_buf[7];
	            vin_string[15] =  j1850_disp_buf[8];
	            vin_string[16] =  j1850_disp_buf[9];
		    	    break;		    	    		    	    		    	                 
              }
             vin_packet++; 
            }
        }
   return vin_packet;     
  }

#if 0
// Display VIN responce packets
// LCD cursor position must be set elsewhere
void display_vin(void)
  {
   if(eof)
        {
        eof = 0;                                  // buffer used
        if((j1850_disp_buf[0] == 0x48) && (j1850_disp_buf[3] == 0x49)) // display filter, only diag MODE 09 responce messages 
            {
             lcd_putch(j1850_disp_buf[6]); 
             lcd_putch(j1850_disp_buf[7]);
             lcd_putch(j1850_disp_buf[8]);
             lcd_putch(j1850_disp_buf[9]);        
            }
        }
  }
# endif


// Compare display filter, 0x00 filter = don't care

char compare_disp_filter()     // 0=no match, !0=match
  {
   char x;
   
   for(x=0; x<DISP_FILTER_LEN; x++)
    {
     if(disp_filter[x] != 0)
      {
       if(disp_filter[x] != j1850_disp_buf[x])
          return 0;                             // no match
      }
    }
   return 1;                                    // match
  }


// Raw diag data display

void raw_diag_display(void)
  {
  char x;
  
//  if(j1850_disp_buf[0] == 0x48)                // display filter, only diag responce messages   
//  if( 1 )   // test 
    if(compare_disp_filter())                    // compare display filter
      {  
      set_cur_lcd(LINE1);
      
       for(x=0; x<9; x++)                        // display part (9-bytes) of RX buffer
        {
        if(x < j1850_disp_bytes)
          lcd_putchhex( j1850_disp_buf[x] );
        else
          {
          lcd_putch('_');
          lcd_putch('_');
          }
        lcd_putch(' ');                           // byte separator
        
        if(x == 4)                                // 5 bytes / line
          {
          lcd_putch(' ');                         // space stuff
          set_cur_lcd(LINE2);                     // next line
          }
        }
      
      lcd_putch(' ');
      lcd_putch(' ');
 
      lcd_putchhex(j1850_disp_bytes);             // display RX characters
      lcd_putch(' ');
      }
 
      
    if((SW12_PORT & SW2) == 0)                    // addjust first display filter digit "on the fly"
     {
      disp_filter[0]--; 
      ms_delay(100);                              // debounce / repeat delay
     }      
   
   if((SW34_PORT & SW3) == 0)
     {
      disp_filter[0]++; 
      ms_delay(100);                              // debounce / repeat delay
     }     
      
  }

// Formated diag data display

void formated_diag_display(void)
  {
//  unsigned int temp;
   char iat, tmp, adv;

   if(j1850_disp_buf[0] == 0x48)                // display filter, only diag responce messages
      {
      
// display line 1      
      if(j1850_disp_buf[4] == 0x0F)             // IAT
          {              
          set_cur_lcd(LINE1);
          lcd_putch('t');
          lcd_putch(' ');
          
          iat = j1850_disp_buf[5] - 40;           // scale IAT
          lcd_put_3_int(iat);                     // display IAT
          }
          
 
      if(j1850_disp_buf[4] == 0x05)             // Coolant Temp
          {              
          set_cur_lcd(LINE1+5);
          lcd_putch(' ');
          tmp = j1850_disp_buf[5] - 40;           // scale Temp          
          lcd_put_3_int(tmp);                     // display Temp
          }
          
     if(j1850_disp_buf[4] == 0x03)             // Fuel system status
          {              
          set_cur_lcd(LINE1+9);
          lcd_putch(' ');
          lcd_putch(' ');
          if((j1850_disp_buf[5] && 0x02) != 0)  // closed loop
              lcd_putch('C');     // closed
          else
              lcd_putch('O');     // open
              
          lcd_putch('L');
          lcd_putch(' ');
          lcd_putchhex(j1850_disp_buf[5]);      // show also raw data
          lcd_putch(' ');    
          
          }
                    
                        
// display line 2     

      if(j1850_disp_buf[4] == 0x0B)             // MAP
          {              
          set_cur_lcd(LINE2);
          lcd_putst("MAP ");
          lcd_putu_3_char(j1850_disp_buf[5]);  // 1:1 scaling 
          lcd_putch(' ');
          }
#if 0
      if(j1850_disp_buf[4] == 0x11)             // TPS
          {              
          set_cur_lcd(LINE2);
          lcd_putst("TPS ");
          lcd_putu_3_char(j1850_disp_buf[5]);  // 1:1 scaling 
          lcd_putch(' ');
          }
         
           
      if(j1850_disp_buf[4] == 0x0C)             // RPM
          {              
          set_cur_lcd(LINE2+8);
          lcd_putst("RPM ");
          temp = (((unsigned int)j1850_disp_buf[5]<<8) + j1850_disp_buf[6]);  // scale RPM
          temp = temp >> 2;                    // divide by four
          lcd_putu_4_int(temp);
          }
#endif          
           
      if(j1850_disp_buf[4] == 0x0E)             // Ign ADV
          {              
          set_cur_lcd(LINE2+8);
          lcd_putst(" ADV ");
          adv = j1850_disp_buf[5] >> 1;           // scale +/- 64 deg
          adv = adv - 64;                         // offset
          lcd_put_3_int(adv);                     // Display
          }              
                    
      }

  }
 
// Display some Jeep protocol messages

// filter and display messages 0x10, 0x37, 0x3A
// transmission in/out rpm, MAP
// transmission gear status & shifter position

void disp_Jeep_1(void)
  {
  unsigned int temp;
  
  if(j1850_disp_buf[0] == 0x10)                //filter 0x10 RPM & MAP message
      {
// display line 1        
      set_cur_lcd(LINE1);
      lcd_putst("In ");
      temp = (((unsigned int)j1850_disp_buf[1]<<8) + j1850_disp_buf[2]);  // scale input RPM
      temp = temp >> 2;                    // divide by four
      lcd_putu_4_int(temp);   
 
      lcd_putst(" Out ");
      temp = (((unsigned int)j1850_disp_buf[3]<<8) + j1850_disp_buf[4]);  // scale output RPM
      temp = temp >> 2;                    // divide by four
      lcd_putu_4_int(temp);
      
// display line 2      
      set_cur_lcd(LINE2+13);
      lcd_putu_3_char(j1850_disp_buf[5]);  // 1:1 scaling  for MAP     
      }
      
  if(j1850_disp_buf[0] == 0x37)                //filter 0x37 Shifter position message
      {      
      set_cur_lcd(LINE2);
      lcd_putst("Gear ");
      
      switch(j1850_disp_buf[1])
        {
        case 1:
         lcd_putch('P');
			  break;
			  
        case 2:
         lcd_putch('R');
			  break; 			   

        case 3:
         lcd_putch('N');
			  break;
			  
        case 5:
         lcd_putch('D');
			  break; 			   

        case 6:
         lcd_putch('2');
			  break;
			  
        case 7:
         lcd_putch('1');
			  break; 			   

        default:
         set_cur_lcd(LINE2+4);
         lcd_putchhex(j1850_disp_buf[1]);   // disp hex if unknown status
			  break;   
        }
      }
      
 if(j1850_disp_buf[0] == 0x3A)                //filter 0x3A Transmission status message      
     {      
     set_cur_lcd(LINE2+6);
     lcd_putch(' ');
     
     if(j1850_disp_buf[1] & 0x10)           // 01 R
         lcd_putch('R');
     else
          {
           if(j1850_disp_buf[1] && 0x20)    // 10 F 
                lcd_putch('F');
           else                             // 00 N or not known
                lcd_putch('N');
          }
          
      lcd_putch(' ');
      lcd_putch((j1850_disp_buf[1] &0x0F) + 0x30); // print selected Gear
 
     
      if(j1850_disp_buf[1] & 0x40)          // Partial lock
         lcd_putch('P');
      else
         lcd_putch(' ');
      
   
      if(j1850_disp_buf[1] & 0x80)          // Full lock
         lcd_putch('F');
      else
         lcd_putch(' '); 
     }
  }
  

// filter and display messages 0x80, 0x72, 0xA4, 0xA0
// clock, odometer
// fuel tank%, operating distance


 void disp_Jeep_2(void)
  {
  unsigned int temp;
  unsigned long ltemp;
 
 // display line 1 
  if(j1850_disp_buf[0] == 0x80)                //filter 0x80 clock
      {
      set_cur_lcd(LINE1);
      lcd_putchhex(j1850_disp_buf[1]);      // hr, format = BCD digits, use HEX printout
      lcd_putch(':');
      lcd_putchhex(j1850_disp_buf[2]);      // min
      lcd_putch(' ');      
      }
      
   if(j1850_disp_buf[0] == 0x72)                //filter 0x72 odometer
      {
      set_cur_lcd(LINE1+6);     
      ltemp = j1850_disp_buf[1];                // get odometer binary value
      ltemp = (ltemp << 8) + j1850_disp_buf[2]; // odometer units are 1/8000 miles !
      ltemp = (ltemp << 8) + j1850_disp_buf[3];
      ltemp = (ltemp << 8) + j1850_disp_buf[4];
      ltemp = ltemp / 4971UL;                   // 8000 / 1.609 = 4971
      lcd_putulong(ltemp);
      lcd_putst(" km");
      }

// display line 2
   if(j1850_disp_buf[0] == 0xA4)                //filter 0xA4 fuel tank %
      {      
      set_cur_lcd(LINE2);
      lcd_putst("Fuel ");
      temp = j1850_disp_buf[1];
      temp = (temp * 100) >> 8;                 // scale 100/256 %
      lcd_put_3_int((int)temp);
      lcd_putst("% ");
      }

   if(j1850_disp_buf[0] == 0xA0)                //filter 0xA4 operating distance( 0.1 miles / bit )
      {      
      set_cur_lcd(LINE2+10);
      temp = j1850_disp_buf[1];
      temp = (temp << 8) + j1850_disp_buf[2];   // scaling 0.1miles/bit
//      temp = temp / 10;                         // scaling to miles
      ltemp = ((long)temp * 165L) >> 10;        // scaling to km, dist*1.069*1024/10/1024
      temp = ltemp;
      lcd_put_3_int(temp);
      lcd_putst(" km");
      }
  }


// filter and display messages 0x1A, 0x52
// cruide control packet/data
// AC status


 void disp_Jeep_3(void)
  {

 // display line 1 
  if(j1850_disp_buf[0] == 0x1A)                //filter 0x1A cruise
      {
      set_cur_lcd(LINE1);
      lcd_putst("Tps"); 
      lcd_put_3_int((int)j1850_disp_buf[1]);   // TPS relative value 0-100%
      lcd_putst("%  Set ");
      lcd_put_3_int((int)j1850_disp_buf[2]);   // Cruise speed set value, units = ?

 // display line 2
      set_cur_lcd(LINE2);
      if(j1850_disp_buf[3] & 0x04)             // brake
        lcd_putst("Brake");
      else
        lcd_putst("     ");
      
      set_cur_lcd(LINE2+7);
      if(j1850_disp_buf[3] & 0x02)             // Cruise on/off
        lcd_putst("Cruise");
      else
        lcd_putst("      ");
      lcd_put_3_int((int)j1850_disp_buf[4]);   // unknown parameter
      }
      
  if(j1850_disp_buf[0] == 0x52)                //filter 0x52 AC status
      {
      set_cur_lcd(LINE2+5);
      if(j1850_disp_buf[1] == 0)
         lcd_putst("  "); 
      else
          {
          if(j1850_disp_buf[1] == 0x01)        // AC on  
            lcd_putst("AC");
          else
            lcd_putchhex(j1850_disp_buf[1]);  // not yet known bits, print hex value
          }
      }
      
  }
  
  
void disp_Jeep_4(void)
  {
   int temp;
        
 // display line 1 
  if(j1850_disp_buf[0] == 0x7E)               //filter 0x7E ignition on hours
      {
      set_cur_lcd(LINE1);
      lcd_putst("hr "); 
      lcd_putu_2_char(j1850_disp_buf[1]);     // Hour digits are mod 100
      lcd_putch(':');
      lcd_putu_2_char(j1850_disp_buf[2]);     // Min digits binary coded
      lcd_putch(':');
      lcd_putu_2_char(j1850_disp_buf[3]);     // Seconds digits binary coded
      lcd_putst("    ");       
      }
      
 // display line 2
   if(j1850_disp_buf[0] == 0xEA)              //filter 0xEA Battery possible voltage ???
      {
      set_cur_lcd(LINE2);
      lcd_putst("0xEA ? ");
      temp = j1850_disp_buf[1]; 
      lcd_put_3_int(temp);               //      
      lcd_putch(' ');
      lcd_put_3_int((temp - 40));                 //
      lcd_putst(" ");
      }
  }   
 
 
 void disp_Jeep_5(void)
  {
   char *ptr;
        

  if(j1850_disp_buf[0] == 0xF0)               //filter 0xF0 VIN number
      {
      switch (j1850_disp_buf[1])              // copy VIN number segment to vin_string[]
			  {
			  case 0x01:
         vin_string[0] = j1850_disp_buf[2]; 
        break;
 
 			  case 0x02:
         vin_string[1] = j1850_disp_buf[2];
         vin_string[2] = j1850_disp_buf[3]; 
         vin_string[3] = j1850_disp_buf[4]; 
         vin_string[4] = j1850_disp_buf[5]; 
        break;     	
 
  			case 0x06:
         vin_string[5] = j1850_disp_buf[2];
         vin_string[6] = j1850_disp_buf[3]; 
         vin_string[7] = j1850_disp_buf[4]; 
         vin_string[8] = j1850_disp_buf[5]; 
        break;        	
 
   			case 0x0A:
         vin_string[9] = j1850_disp_buf[2];
         vin_string[10] = j1850_disp_buf[3]; 
         vin_string[11] = j1850_disp_buf[4]; 
         vin_string[12] = j1850_disp_buf[5]; 
        break;
        
        case 0x0E:
         vin_string[13] = j1850_disp_buf[2];
         vin_string[14] = j1850_disp_buf[3]; 
         vin_string[15] = j1850_disp_buf[4]; 
         vin_string[16] = j1850_disp_buf[5]; 
        break;    
			  }
			  
     vin_string[17] = 0;
     ptr = &vin_string[1];
     
 // display line 1 
      set_cur_lcd(LINE1);
      lcd_putst("VIN =          ");
      lcd_putch(vin_string[0]);             // display first digit at line 1
      
 // display line 2
      set_cur_lcd(LINE2); 
      lcd_putst(ptr);                       // display rest of the string at line 2

      }
  }   
   
   

// wait for diag packet responce & display results in raw format

void check_diag_responce(void)                  // watch diag command results & display
    {
    if(eof)
      {
       eof = 0;                                 // buffer used  
       raw_diag_display();                      // display what we got
      }
    }


// Simple edit for display data filter

void edit_display_filter()
  {
   set_cur_lcd(LINE1);
   lcd_putst("Filter digit# ");
   lcd_putchhex(disp_filt_digit); 
   
   set_cur_lcd(LINE2);
   lcd_putst("Filter data   ");  
   lcd_putchhex(disp_filter[disp_filt_digit]); 
  
   if((SW12_PORT & SW1) == 0)
     {
      disp_filt_digit++;
      if(disp_filt_digit > (DISP_FILTER_LEN-1))
         disp_filt_digit = 0;
      ms_delay(200);                     // debounce / repeat delay
     }
 
   if((SW12_PORT & SW2) == 0)
     {
      disp_filter[disp_filt_digit]--; 
      ms_delay(100);                     // debounce / repeat delay
     }      
   
   if((SW34_PORT & SW3) == 0)
     {
      disp_filter[disp_filt_digit]++; 
      ms_delay(100);                     // debounce / repeat delay
     }   
    
  }
  
//
// Display intensity, PWM3

// Init pwm generator
void init_lcd_pwm(void)
  {
  PWMPRCLK = 0x50;                      // clk B = bus clock / 32
  PWMPOL = 0x08;                        // PWM3 polarity = high first
  PWMPER3 = 0xFF;                       // set period, 8MHz/32/256 = 976Hz
  PWMDTY3 = 0x80;                       // set 50% duty cycle
  PWME = 0x08;                          // enable PWM3  
  }


// Set intensity(pwm), higher value = more intensity
// 0 = off, 255 = max
void set_lcd_pwm(unsigned char pwm)
  {
  PWMDTY3 = pwm;                        // set pwm
  }
  
// decode / set LCD intensity from instrument cluster & light status packet
// packet #0x60
// example 60 FF 80 C6 = lights off, max bl intensity
void decode_intensity(void)                             // decode & set LCD back light intensity  
    {
    unsigned char intensity;

    intensity = 0;
    if(j1850_disp_buf[2] & 0x80)                        // evaluate MSB
      intensity = 0x10;  
    
    intensity = intensity | (j1850_disp_buf[1] & 0x0F); // get & add low bits

// after this point possible intensity values are (min to max)
// 01, 03, 06, 09, 0C, 0F, 1F

    switch (intensity)                                  // set LCD back light PWM, values are "from the hat" 
			  {
			  case 0x01:
         lcd_pwm = 10; 
        break;

			  case 0x03:
         lcd_pwm = 20; 
        break;
        
  			case 0x06:
         lcd_pwm = 40; 
        break;
        
 			  case 0x09:
         lcd_pwm = 80; 
        break;
        
 			  case 0x0C:
         lcd_pwm = 140; 
        break;
        
 			  case 0x0F:
         lcd_pwm = 170; 
        break;
        
			  case 0x1F:
         lcd_pwm = 255; 
        break;
			  }
			  
		set_lcd_pwm(lcd_pwm);                              // set back light value	  
  
    }

// ________________________________ Main forever loop ______________________________

void main(void)
 {
  unsigned char x=0;
  unsigned char u, c;

 
  DDRA = DDRA_INIT;                   // Init I/O ports
  DDRB = DDRB_INIT;
   
  PORTA = PORTA_INIT;
  PORTB = PORTB_INIT;
  
  DDRT = DDRT_INIT;
  PTT = PTT_INIT;
  
  DDRP = DDRP_INIT;
  PTP = PTP_INIT;
  
  DDRM = DDRM_INIT;
  PTM = PTM_INIT;


	initlcd();                          // init LCD & say hello & show version
	lcd_putst(msg1);
	lcd_putst(ver);	
	set_cur_lcd(LINE2);
	lcd_putst(msg2);
	ms_delay(3000);
	
	init_lcd_pwm();
	lcd_pwm = 128;
	set_lcd_pwm(lcd_pwm);
	
	SCI0Init();                         // init SCI0 uart & say hello
	putstr("\n\rJ1850 converter test version ");
  putstr((signed char*)ver);
	putstr("\n\r            Juha Niinikoski\n\r");
	
	init_J1850();                       // init BDLC controller
	putstr("\n\rBDLC init done");

  absoluteTime = 0;                   // start 1ms tick
  RTIInit();                          // Enable interrupts

  HPRIO = 0xC2;                       // give BDLC interrupt priority over the others
  EnableInterrupts;
  putstr("\n\rTimer tick & interrupts enabled, HPRIO = ");
  putchhex(HPRIO);
  
  for(x=0; x<17; x++)                 // try to read VIN
      {
       vin_string[x] = '?';
      }
  vin_string[17] = 0;      
  vin_packet = 0;                     // prepare read
  J1979_send(MODE9, 0x02);            //ask VIN number, MODE 09, Info Type 02
  ms_timer = 500;
  
  while(ms_timer != 0 && read_vin() < 5); // try to read VIN or leave with timeout
  
  putstr("\n\rHost VIN# from ECU # ");
  putchhex(vin_ecu);
  putstr(" = ");
  putstr(vin_string);
  putstr("\n\r");
     
  
 // main forever loop 
 
  for(;;)
   {
  
  // all sort of J1850 tests
  
   c = kbhit();                                 // check PC keyboard input
   if (c != 0)
		{
    c = getch();                                // get command character
		switch (c)
			{
			case 'l':
		    DLCBCR2 = DLCBCR2 ^ 0x40;               // toggle loop back
		    print_reg();
			break;
   
 			case 'p':
		    print_reg();
			break;
   
 			case 't':
        putstr("\n\rTX J1850, give data ");
        for(x=0; x<J1850_TX_BUF_LEN; x++)      // collect tx data to buf len
          {
           putstr("\n\rTX buffer[");
           putchhex(x);
           putstr("] "); 
          u = getch();
          putch(u);
          if(u == '\n' | u == '\r')           // or  to cr or lf
            break;
          
          j1850_tx_buf[x] = u;
          }
          
        if(x == 0)
          {
           putstr("\n\rNo input data");
           break;
          }
          
        j1850_tx_bytes = x;                   // prepare trensmission
        tdre_cnt = 0;

//        j1850_rx_bytes = 0;                   // clear RX buffer pointer
        for(x=0; x<J1850_RX_BUF_LEN; x++)
            j1850_rx_buf[x] = 0;
          
        DLCBDR = j1850_tx_buf[tdre_cnt];      // load tx register, tx first byte
        tdre_cnt++;

//       tx_j1850_buffer(j1850_tx_bytes);
//		    print_reg();
			break;

			case 'b':                                 // check uart baud rate
           putstr("\n\rBaud rate SCI0BDL = ");
           putchhex(SCI0BDL);

			break;

			case '+':
			  lcd_pwm++;
		    set_lcd_pwm(lcd_pwm);
			break;
			
			case '-':
			  lcd_pwm--;
		    set_lcd_pwm(lcd_pwm);
			break;
						
			 default:
           putchhex(c);
           putstr(" = ? cmd\n\r");              // unknown command
   
			}   // end switch
		}     // end if	  


// Automatic send for diag commands, 100ms / command

    if(disp_page == 0)                               // send commands only when display selected 
        send_diag_commands();

 
// LCD display tests & serial line log transmit

    if(eof)
        {
        if(j1850_disp_buf[0] == 0x60)                // body controller light status message
          decode_intensity();                        // decode & set LCD back light intensity
        
 		    switch (disp_page)                           // select LCD display page
			    {
			    
// formated displays
			    
		    	case 0:                                    // send OBD-II command & display results
			      formated_diag_display();                 // formated OBD-II test display
		    	break;
    
   	
		    	case 1:                                    // display Jeep messages set #1
			     disp_Jeep_1();
		    	break;

		    	
		    	case 2:                                    // display Jeep messages set #2
			     disp_Jeep_2();
		    	break;

		    	
	    	  case 3:                                    // display Jeep messages set #3
			     disp_Jeep_3();
		    	break;		    			    			    	


// raw data test displays
		    	case 4:                                    // edit display/capture filter
			     edit_display_filter();
		    	break;
		    	 

	    	  case 5: default:                           // raw display
			     raw_diag_display();
		    	break; 
 

 		    	case 6:                                    // data log printout
 		    	 print_j1850_rx(j1850_disp_bytes);         // put record in log print queue
 		    	 set_cur_lcd(LINE2);
 		    	 lcd_putst("Log msg# ");
 		       lcd_putulong(rx_counter);                        
//			     raw_diag_display();                     // raw display
		    	break;

    	  case 7:                                      // display Jeep messages set #4, test
			      disp_Jeep_4();
		    	break; 
 
 	
	   	  case 8:                                      // display Jeep messages set #5, VIN
			      disp_Jeep_5();
		    	break; 
		    	

			    }       // end switch
			    
		    eof = 0;                                      // buffer used	    
        }         // end if
        
      
        

 
    
// Handle user interface push button switches

// send some test commands


  if(disp_page != 4)                       // buttons have different function in filter edit
    {
      
#if 0
    if((SW12_PORT & SW1) == 0)
      {
       clear_lcd();      
       lcd_putst("CMD 01 0B MAP   ");
       J1979_send(MODE1, 0x0B);            //ask MAP

       while((SW12_PORT & SW1) == 0)
          {  
          check_diag_responce();             // display responce if any
          }      
       }
#endif 
 
     if((SW12_PORT & SW1) == 0)
      {
       clear_lcd();      
       lcd_putst("Read VIN ");
       
           
       vin_packet = 0;                     // prepare read
       vin_string[17] = 0;
       J1979_send(MODE9, 0x02);            //ask VIN number, MODE 09, Info Type 02

       while(((SW12_PORT & SW1) == 0) && (read_vin() < 5)); // try to read 5 packets
       
       
       lcd_putchhex(vin_ecu);              // display responding module
       set_cur_lcd(LINE1+15);
       lcd_putch(vin_string[0]);           // display first digit
       
       set_cur_lcd(LINE2);  
       for(x=1; x<17; x++)                 // display sest of the digits
            {
             lcd_putch(vin_string[x]); 
            }
            
       while((SW12_PORT & SW1) == 0);      // keep display while button pressed  
       }
      
      
    if((SW12_PORT & SW2) == 0)          // test
      {
       clear_lcd();
       lcd_putst("CMD 01 04 Load  "); 
       J1979_send(MODE1, 0x04);             //  ask Load 
              
       while((SW12_PORT & SW2) == 0)
          {  
           check_diag_responce();           // display responce if any
          }       
      }
      
      
    if((SW34_PORT & SW3) == 0)          // test
      {
    
       clear_lcd();
       lcd_putst("CMD 01 0E Adv. ");       
       J1979_send(MODE1, 0x0E);             // ask Ignition advance

       while((SW34_PORT & SW3) == 0)
          {  
          check_diag_responce();             // display responce if any
          }           
      }
      

    }       // end if


// select display / operating mode
      
    if((SW34_PORT & SW4) == 0)
      { 
       print_statistic();                 // print BDLC statistics to serial port
       
       disp_page++;                       // addjust display page
			 if(disp_page > (DISP_PAGES-1))
			    disp_page = 0;
			 
			 putstr("\n\rDisplay page = ");
       putchhex(disp_page);

       clear_lcd();       
       lcd_putst("Disp. Page = ");
       lcd_putchhex(disp_page);
       
       while((SW34_PORT & SW4) == 0)      // you can addjust back light when SW4 pressed
          {
           if((SW12_PORT & SW2) == 0)     // darker
              {
              if(lcd_pwm != 0)
                lcd_pwm--;
              ms_delay(100); 
              }
              
           if((SW34_PORT & SW3) == 0)     // brighter
              {
              if(lcd_pwm < 255)
                lcd_pwm++;
              ms_delay(100);
              }
              
          set_cur_lcd(LINE2);       
          lcd_putst("Back Light = ");     // show & set pwm value
          lcd_put_3_int((int)lcd_pwm);
          set_lcd_pwm(lcd_pwm);
          
          }
       ms_delay(200);                     // debounce
      }



//
// carbage test code storage
//_____________________________________________________________________________

#if 0 

// switch tests   
    if((PORTB & 0x20) == 0)  
      PORTA = PORTA | 0x02;
    else
      PORTA = PORTA & ~0x02;
   
 
    if((SW12_PORT & SW1) == 0)  
      PORTA = PORTA | 0x01;
    else
      PORTA = PORTA & ~0x01;
    
    
      if((SW34_PORT & SW3) == 0)
      {
       clear_lcd();
       lcd_putst("Switch 3");
       while((SW34_PORT & SW3) == 0)
        {
        set_cur_lcd(LINE1);
        lcd_putchhex(kbhit());      // test, kb input buffer manager
        }       
      }



// sprintf test    
// lcdbuff_len =   sprintf(lcdpbuff, " = %X, l =%i  ",x ,lcdbuff_len );
//	lcd_putbuf(lcdpbuff, lcdbuff_len);

// LCD speed test   
   clear_lcd();
   while(1)  // LCD disp speed test
    {        // about 2ms / full round
    i++;
    c++;
      
    set_cur_lcd(LINE1);
     
     lcd_putu_2_char(c);
     lcd_putch(' ');
     
     lcd_putu_3_char(c); 
     lcd_putch(' ');
     
     lcd_putchhex(c);
     lcd_putch(' ');
     
     lcd_put_3_int((unsigned int)i);
     lcd_putch(' ');
  
    set_cur_lcd(LINE2);
       
     lcd_putu_4_int(i);
     lcd_putch(' ');
     
     lcd_putulong(i);
     lcd_putch(' '); 
     
     while((SW12_PORT & SW1) == 0);
     ms_delay(100);
    }
   


// BDLC I/O  
  
static void tx_j1850_buffer(unsigned char amount) // transmit n characters from buffer
  {
  char x;
 
  
  for(x=0; x<amount; x++)
    {
      
    DLCBDR = j1850_tx_buf[x];       // load tx register
//    DLCBCR1_IMSG = 1;               // defeat receiver
    
    if(x == (amount-1))             // if last byte flag it
      {    
      DLCBCR2_TEOD = 1;
      break;                        // leave tx routine if transmit last
      }
      
  
    
    while(DLCBSVR < 0x10);          // wait for BDLC TX status

    if(DLCBSVR != 0x10)              // We accept only Tx empty ?
      {
      putstr("\n\rTX error");
      break;
      }
    }
  
  }      
    
#endif      
 
 
   
   }   // end for(;;)
}      // end main

       /* please make sure that you never leave this function */