//
// Simple LCD driver for 9s12DJ64 controller
//
// Juha Niinikoski 22.05.2007
//
// LCD printout speed optimization 31.05.2007
//


#include <hidef.h>                        /* common defines and macros */
#include <mc9s12dj64.h>                   /* derivative information */
#pragma LINK_INFO DERIVATIVE "mc9s12dj64"

#include "J1850_conv.h"                   // board specific definitions


char lcdpbuff[20];                        // sprintf buffer for LCD printouts
int lcdbuff_len;                          // buffer fill length


// Micro second delay, only for 16MHx Xtal/8MHz E-clock, not exactly accurate

void us_delay(unsigned int delay)         // microsedond delay
  {
  unsigned int x;
  
  delay = delay >> 1;
  for(x=0; x<delay; x++)
    {
    asm ("nop");   
    }
  }
  
// Millisecond delay  
void ms_delay(unsigned int delay)         // millisecond delay 
  {
  unsigned int x;
  
  for(x=0; x<delay; x++)
    {
    us_delay(1000);   
    }
  }


// LCD driver

// External functions
//extern void us_delay( unsigned int);
//extern void ms_delay( unsigned int);


void	clk_lcd(unsigned char ldata)		// set data & clock both lcd controllers
	{
	LCD_DATA = ldata;                 	// set LCD data lines
	asm ("nop");							          // extra delay for filter board latch wires
	LCD_CONTROL = LCD_CONTROL | LCD_E;
	asm ("nop");
	LCD_CONTROL = LCD_CONTROL & ~LCD_E;
	asm ("nop");
	}

void set_cur_lcd(unsigned char cur)	  	// set cursor position
	{
	LCD_CONTROL = LCD_CONTROL & ~LCD_RS;	// switch command mode		
	clk_lcd(cur | CURSET);				        // set cursor command + cursor position
	LCD_CONTROL = LCD_CONTROL | LCD_RS; 	// back to data mode
//	ms_delay(2);
  us_delay(40);                         // nominal should be 37us
	}


void clear_lcd(void)					          // clear display
	{
	LCD_CONTROL = LCD_CONTROL & ~LCD_RS;	// switch command mode		
	clk_lcd(CLEAR);					            	// clear display. Display going to be busy 1,6 mS
	LCD_CONTROL = LCD_CONTROL | LCD_RS;		// back to data mode
	ms_delay(2);                          // should me 1.6ms
	}


void initlcd(void)						          // init LCD controller
	{

	LCD_CONTROL = LCD_CONTROL & ~LCD_RS;	// set command mode
	LCD_CONTROL = LCD_CONTROL & ~LCD_RW;  // set write direction
	ms_delay(15);						              // start delay
	clk_lcd(0x38);						            // write init data with delays
	ms_delay(5);
	clk_lcd(0x38);
	ms_delay(1);
	clk_lcd(0x38);						            //8-bit, 2 rows, 5x7 matrix, display on
	ms_delay(1);						              //cursor off, no blink
	clk_lcd(0x0C);
	ms_delay(1);
	clk_lcd(0x01);
	ms_delay(2);	
	clk_lcd(0x06);						            // increment, no display shift
	ms_delay(1);
	LCD_CONTROL = LCD_CONTROL | LCD_RS;		// set data mode
	
	LCD_CONTROL = LCD_CONTROL | LCD_BL;   // turn back light on
	}


void lcd_putch(char c)                  // print one character
  {
  clk_lcd(c);
  us_delay(40);                         // write with wait loop, should be 37us  
  }
  
void lcd_putst(const char *str)         // print null terminated string
	{
	while((*str)!=0)
		{
		lcd_putch(*str);
		str++;
		}
	}

void lcd_putbuf(char *str, int len)     // print n character string
	{
	while(len > 0)
		{
		lcd_putch(*str);
		str++;
		len--;
		}
	}

#if 0	
void lcd_putchhex(unsigned char c)        // print HEX character
{
	unsigned char temp;

	temp=c;

	c=(c >> 4);
	if (c<10) c+=48; else c+=55;
	lcd_putch(c);

	c=temp;

	c=(c & 0x0F);
	if (c<10) c+=48; else c+=55;
	lcd_putch(c);
}	
#endif


void puthexnible( unsigned char c )       // print HEX nibble
	{
	c=(c & 0x0F);					// mask high nibble out
	if (c<10) c+=48; else c+=55;	// HEX convert
	lcd_putch(c);					// display
	}


void lcd_putchhex(unsigned char c)          // print HEX character
{
	unsigned char cl;

	cl=c;
	c=(c >> 4);						// shift high nibble down
	puthexnible(c);					// display high nibble
	puthexnible(cl);				// display low nibble
}



void lcd_putlonghex( long ww )                // HEX long
	{

	union
	{
	long w;
	unsigned char p[4];
	} lrek;

	lrek.w = ww;
	lcd_putchhex(lrek.p[0]);
	lcd_putchhex(lrek.p[1]);
	lcd_putchhex(lrek.p[2]);
	lcd_putchhex(lrek.p[3]);
	}


void bin2ascii_l(unsigned long luku)           // convert 32-bit number to ASCII string
	{
char x, idx;

	idx = 6;                  // result index[0]...[6]
	for(x=0; x<7; x++)				// 7 digits, 00.000.000
		{
		lcdpbuff[idx] = (luku % 10) + '0';
		luku = luku / 10;
		idx--;
		}
	lcdpbuff[7] = 0;				// string terminator
	}


void bin2ascii_ui(unsigned int luku)           // convert 16-bit number to ASCII string
	{
char x, idx;

	idx = 6;                  // result index[2]...[6]
	for(x=0; x<5; x++)				// 5 digits, 00.000
		{
		lcdpbuff[idx] = (luku % 10) + '0';
		luku = luku / 10;
		idx--;
		}
	lcdpbuff[7] = 0;				// string terminator
	}


void bin2ascii_c(unsigned char luku)           // convert 8-bit number to ASCII string
	{
char x, idx;

	idx = 6;                  // result index[4]...[6]
	for(x=0; x<3; x++)				// 3 digits, 000
		{
		lcdpbuff[idx] = (luku % 10) + '0';
		luku = luku / 10;
		idx--;
		}
	lcdpbuff[7] = 0;				// string terminator
	}


void lcd_putulong( unsigned long l)           // print 7-digit long number
	{
	bin2ascii_l(l);					      // partial convert long number to string

	lcd_putch(lcdpbuff[0]);			// 1 000 000
	lcd_putch(lcdpbuff[1]);			// 100 000
	lcd_putch(lcdpbuff[2]);			// 10 000
	lcd_putch(lcdpbuff[3]);			// 1 000
	lcd_putch(lcdpbuff[4]);			// 100
	lcd_putch(lcdpbuff[5]);			// 10
  lcd_putch(lcdpbuff[6]);     // 1
	}
	
	
void lcd_putu_4_int(unsigned int ui)          // print 4-digit int number
	{
	if(ui > 9999)               // limit to 4 digits
	  ui = 9999;
	
	bin2ascii_l(ui);					  // convert long number to string

	lcd_putch(lcdpbuff[3]);			// 1 000
	lcd_putch(lcdpbuff[4]);			// 100
	lcd_putch(lcdpbuff[5]);			// 10
  lcd_putch(lcdpbuff[6]);     // 1
	}


void lcd_put_3_int( int c )                  // print 3-digit int number -99...999
	{
	char n;
	n = 0;
	if(c > 999)
	  c = 999;
	if(c < -99)
	  c = -99;
	
  if(c < 0)                                   // test if negative
      {
       n = 1;
       c = 0 - c;
      }
      
	bin2ascii_ui((unsigned int)c );					  //  convert number to string
  if(n == 0)
    {
     if(lcdpbuff[4] == '0')
       lcd_putch(' ');
     else
	     lcd_putch(lcdpbuff[4]); // 100
    }
	else
	  lcd_putch('-');
	
	lcd_putch(lcdpbuff[5]);			// 10
  lcd_putch(lcdpbuff[6]);     // 1
	}


void lcd_putu_3_char( unsigned char c)        // print 3-digit char number
	{
	bin2ascii_c( c );					  //  conver number to string

	lcd_putch(lcdpbuff[4]);			// 100
	lcd_putch(lcdpbuff[5]);			// 10
  lcd_putch(lcdpbuff[6]);     // 1
	}


 void lcd_putu_2_char( unsigned char c)        // print 2-digit char number
	{
	 if(c > 99)
	  c = 99;
   lcd_putch((c / 10) + '0');               // 10
   lcd_putch((c % 10) + '0');               // 1
	}
