//
// Simple Uart0 driver for 9s12DJ64 controller
//
// Juha Niinikoski 22.05.2007
//
// IRQ control 23.05.2007
// Mod 256 TX buffer wrap around bug corrected 27.05.2007


#include <hidef.h>                        /* common defines and macros */
#include <mc9s12dj64.h>                   /* derivative information */
#pragma LINK_INFO DERIVATIVE "mc9s12dj64"

#include "J1850_conv.h"                   // board specific definitions


// Module variables
// Buffer pointer logic is based on mod 256 arithmetic. Use only 256 long buffers !

static unsigned char tx_buffer[256];      // SCI0 port TX queue
static unsigned char tx_buf_out_idx = 0;  // buffer output index
static unsigned char tx_buf_in_idx = 1;   // buffer input index

static unsigned char rx_buffer[256];      // SCI0 port TX queue
static unsigned char rx_buf_out_idx = 0;  // buffer output index
static unsigned char rx_buf_in_idx = 1;   // buffer input index

static unsigned char pbuff[8];            // print format buffer


// This code is called from IRQ handler

void sci0_int(void)                        // put receiced character to RX queue
  {
  unsigned char rx;
  
  if(SCI0SR1_RDRF)                          // RX code, allways active
     {  
// Receive character from SCI0 
      rx = SCI0SR1;                        // dummy read to clear RDRF flag
      rx = SCI0DRL;                        // data read
  
      if((rx_buf_in_idx + 1) != rx_buf_out_idx) // check overrun
        {
        rx_buffer[rx_buf_in_idx++] = rx;    // place just received character in queue
        }
        // else reject character
     }
      
  else                                     // TX code active only when stuff in queue
     {
     if(SCI0CR2_SCTIE)                     // run this code only when TIE active
        {          
// send character to SCI0 from TX queue
        if((unsigned char)(tx_buf_out_idx+1) == tx_buf_in_idx) // last character out
          SCI0CR2_SCTIE = 0;                    // stop TX interrupts
          
        SCI0DRL = tx_buffer[tx_buf_out_idx++];  // send character

        }
      }
  }



// Some simple & fast OH2NLT SCI I/O routines


// init SCI0 for IRQ operation
void SCI0Init(void)                       // Init SCI 0
  {
  SCI0BDL = (unsigned char)((16000000UL /* OSC freq */ / 2) / 38400 /* baud rate */ / 16 /*factor*/); 
  SCI0CR2 = 0x2C;                         // RX, TX enable & RX int enable
  }


// put character to TX queue
void putch(unsigned char ch)              /* Print one character */
  {
 
 if((tx_buf_in_idx + 1) != tx_buf_out_idx) // check overrun
    {
    tx_buffer[tx_buf_in_idx++] = ch;
    if(SCI0CR2_SCTIE == 0)                // check if SCI0 "running"
       SCI0CR2_SCTIE = 1;                 // no, enable SCI0 TX interrupts
    }
// else reject character
  
  }


// print hec character
void putchhex(unsigned char c)            /* Print HEX character */
  {
	unsigned char temp;

	temp=c;

	c=(c >> 4);
	if (c<10) c+=48; else c+=55;
	putch(c);

	c=temp;

	c=(c & 0x0F);
	if (c<10) c+=48; else c+=55;
	putch(c);
  }

// print long hex character
void putlonghex(unsigned long ww)        /* Print long  HEX character */
	{

	union
	{
	long w;
	unsigned char p[4];
	} lrek;

	lrek.w = ww;
	putchhex(lrek.p[0]);                    // Motorola byte order
	putchhex(lrek.p[1]);
	putchhex(lrek.p[2]);
	putchhex(lrek.p[3]);
	}


void bin2ascii(unsigned long luku)           // convert number to ASCII string
	{
char x, idx;

	idx = 6;
	for(x=0; x<7; x++)				  // 7 digits, 0.000.000
		{
		pbuff[idx] = (luku % 10) + '0';
		luku = luku / 10;
		idx--;
		}
	pbuff[7] = 0;				        // string terminator
	}


void putulong( unsigned long u)           // print 7-digit long number
	{
	bin2ascii(u);					// partial convert long number to string

	putch(pbuff[0]);			// 1 000 000
	putch(pbuff[1]);			// 100 000
	putch(pbuff[2]);			// 10 000
	putch(pbuff[3]);			// 1 000
	putch(pbuff[4]);			// 100
	putch(pbuff[5]);			// 10
  putch(pbuff[6]);      // 1
 
	}

void putuint(unsigned int i)            // print 5-digit long number
	{
	bin2ascii((unsigned long)i);					// partial convert long number to string

	putch(pbuff[2]);			// 10 000
	putch(pbuff[3]);			// 1 000
	putch(pbuff[4]);			// 100
	putch(pbuff[5]);			// 10
  putch(pbuff[6]);      // 1 
	}


// Print null terminated string
void putstr(char *text)                           // Print string 
  {
  while (*text != '\0')
    {
    putch(*text++);
    }
  }


// Check if characters in queue
unsigned char kbhit(void)                         // check key hit, return number of characters
  {  
  return (rx_buf_in_idx  - (rx_buf_out_idx + 1));   // return wit number of available characters
  }
  
// Get character fron queue
unsigned char getch(void)                         // get one character from rx queue
  {
  while((rx_buf_out_idx + 1) == rx_buf_in_idx);   // wait data if queue is empty
  
  return rx_buffer[++rx_buf_out_idx];             // return with data
  }

